@@ -902,11 +902,11 @@ static int cstp_can_gen_tokencode(struct openconnect_info *vpninfo,
 	}
 #endif
 	/* Otherwise it's an OATH token of some kind. */
-	if (strcmp(opt->name, "secondary_password") &&
-        (!form->auth_id || strcmp(form->auth_id, "challenge")))
-		return -EINVAL;
+	if (!strcmp(opt->name, "secondary_password") ||
+        (form->auth_id && !strcmp(form->auth_id, "challenge")))
+	  return can_gen_tokencode(vpninfo, form, opt);
 
-	return can_gen_tokencode(vpninfo, form, opt);
+	return -EINVAL;
 }
 
 static int fetch_config(struct openconnect_info *vpninfo)
