@@ -879,8 +879,8 @@ JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setCSDWr
 	    !get_cstring(ctx->jenv, jarg2, &arg2)) {
 		openconnect_setup_csd(ctx->vpninfo, getuid(), 1, arg0);
 
-		openconnect_set_csd_environ(ctx->vpninfo, "TMPDIR", arg1);
-		openconnect_set_csd_environ(ctx->vpninfo, "PATH", arg2);
+		if (arg1) openconnect_set_csd_environ(ctx->vpninfo, "TMPDIR", arg1);
+		if (arg2) openconnect_set_csd_environ(ctx->vpninfo, "PATH", arg2);
 	}
 
 	release_cstring(ctx->jenv, jarg0, arg0);
@@ -925,6 +925,12 @@ JNIEXPORT jboolean JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_hasT
 	return openconnect_has_tss_blob_support();
 }
 
+JNIEXPORT jboolean JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_hasTSS2BlobSupport(
+	JNIEnv *jenv, jclass jcls)
+{
+	return openconnect_has_tss2_blob_support();
+}
+
 JNIEXPORT jboolean JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_hasStokenSupport(
 	JNIEnv *jenv, jclass jcls)
 {
@@ -1108,6 +1114,16 @@ JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setXMLPo
 	openconnect_set_xmlpost(ctx->vpninfo, arg);
 }
 
+JNIEXPORT jint JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_getIdleTimeout(
+	JNIEnv *jenv, jobject jobj)
+{
+	struct libctx *ctx = getctx(jenv, jobj);
+
+	if (!ctx)
+		return -EINVAL;
+	return openconnect_get_idle_timeout(ctx->vpninfo);
+}
+
 /* simple cases: return a const string (no need to free it) */
 
 #define RETURN_STRING_START \
@@ -1197,6 +1213,14 @@ JNIEXPORT jstring JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_getCS
 	RETURN_STRING_END
 }
 
+JNIEXPORT jstring JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_getProtocol(
+	JNIEnv *jenv, jobject jobj)
+{
+	RETURN_STRING_START
+	buf = openconnect_get_protocol(ctx->vpninfo);
+	RETURN_STRING_END
+}
+
 #define SET_STRING_START(ret) \
 	struct libctx *ctx = getctx(jenv, jobj); \
 	const char *arg = NULL;			 \
@@ -1247,6 +1271,16 @@ JNIEXPORT jint JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setHTTPP
 	return ret;
 }
 
+JNIEXPORT jint JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setProtocol(
+	JNIEnv *jenv, jobject jobj, jstring jarg)
+{
+	int ret;
+	SET_STRING_START(-ENOMEM)
+	ret = openconnect_set_protocol(ctx->vpninfo, arg);
+	SET_STRING_END();
+	return ret;
+}
+
 JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setXMLSHA1(
 	JNIEnv *jenv, jobject jobj, jstring jarg)
 {
@@ -1263,6 +1297,14 @@ JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setHostn
 	SET_STRING_END();
 }
 
+JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setVersionString(
+	JNIEnv *jenv, jobject jobj, jstring jarg)
+{
+	SET_STRING_START()
+	openconnect_set_version_string(ctx->vpninfo, arg);
+	SET_STRING_END();
+}
+
 JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setUrlpath(
 	JNIEnv *jenv, jobject jobj, jstring jarg)
 {
@@ -1381,3 +1423,53 @@ JNIEXPORT jobject JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_getIP
 
 	return jobj;
 }
+
+JNIEXPORT jobjectArray JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_getSupportedProtocols(
+	JNIEnv *jenv, jclass jcls)
+{
+	jmethodID mid;
+	jobjectArray result;
+	struct libctx ctx = { .jenv = jenv, .jobj = NULL, .async_lock = NULL, .vpninfo = NULL, .cmd_fd = -1, .loglevel = -1 };
+
+	/* call C library */
+	struct oc_vpn_proto *protos;
+	int np, ii;
+	np = openconnect_get_supported_protocols(&protos);
+	if (np < 0)
+		return NULL;
+
+	/* get VPNProto class, its init method, and create array  */
+	jcls = (*jenv)->FindClass(jenv,
+				       "org/infradead/libopenconnect/LibOpenConnect$VPNProto");
+	if (jcls == NULL)
+		goto err;
+	mid = (*jenv)->GetMethodID(jenv, jcls, "<init>", "()V");
+	if (!mid)
+		goto err;
+	result = (*jenv)->NewObjectArray(jenv, np, jcls, NULL);
+	if (result == NULL)
+		goto nomem;
+
+	for (ii=0; ii<np; ii++) {
+		jobject jobj = (*jenv)->NewObject(jenv, jcls, mid);
+		if (!jobj)
+			goto nomem;
+
+		if (set_string(&ctx, jobj, "name",        protos[ii].name) ||
+		    set_string(&ctx, jobj, "prettyName",  protos[ii].pretty_name) ||
+		    set_string(&ctx, jobj, "description", protos[ii].description) ||
+		    set_int   (&ctx, jobj, "flags",       protos[ii].flags))
+			goto nomem;
+
+		(*jenv)->SetObjectArrayElement(jenv, result, ii, jobj);
+	}
+
+	openconnect_free_supported_protocols(protos);
+	return result;
+
+nomem:
+	OOM(jenv);
+err:
+	openconnect_free_supported_protocols(protos);
+	return NULL;
+}
